"""A handler for remote datasets.

DAP handlers convert from different data formats (NetCDF, eg) to the internal
dapclient model. The dapclient client is just a handler that converts from a remote
dataset to the internal model.

"""

import codecs
import copy
import gzip
import io

# handlers should be set by the application
# http://docs.python.org/2/howto/logging.html#configuring-logging-for-a-library
import logging
import pickle
import pprint
import re
import sys
import urllib.parse
from io import BytesIO
from itertools import chain

import numpy

import dapclient.handlers.lib
import dapclient.model
import dapclient.net
import dapclient.parsers
from dapclient.lib import (
    DAP2_ARRAY_LENGTH_NUMPY_TYPE,
    DEFAULT_TIMEOUT,
    START_OF_SEQUENCE,
    BytesReader,
    StreamReader,
    combine_slices,
    encode,
    fix_slice,
    hyperslab,
    walk,
)
from dapclient.parsers.das import add_attributes, parse_das
from dapclient.parsers.dds import dds_to_dataset
from dapclient.parsers.dmr import dmr_to_dataset
from dapclient.responses.dods import DAP2_response_dtypemap

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

BLOCKSIZE = 512


class DAPHandler(dapclient.handlers.lib.BaseHandler):
    """Build a dataset from a DAP base URL."""

    def __init__(
        self,
        url,
        application=None,
        session=None,
        output_grid=True,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_charset="ascii",
        protocol=None,
    ):
        self.application = application
        self.session = session
        self.output_grid = output_grid
        self.timeout = timeout
        self.verify = verify
        self.user_charset = user_charset
        self.url = url

        scheme, netloc, path, query, fragment = urllib.parse.urlsplit(self.url)
        self.scheme = scheme
        self.netloc = netloc
        self.path = path
        self.query = query
        self.fragment = fragment

        self.protocol = self.determine_protocol(protocol)

        self.projection, self.selection = dapclient.parsers.parse_ce(self.query)
        arg = (
            self.scheme,
            self.netloc,
            self.path,
            "&".join(self.selection),
            self.fragment,
        )
        self.base_url = urllib.parse.urlunsplit(arg)
        self.make_dataset()
        self.add_proxies()

    def determine_protocol(self, protocol):
        if protocol == "dap4":
            self.scheme = "http"
            return protocol
        if protocol == "dap2":
            return protocol
        if self.scheme == "dap4":
            self.scheme = "http"
            return "dap4"

        extension = self.path.split(".")[-1]
        if extension in ("dmr", "dap"):
            return "dap4"
        if extension in ("dds", "dods"):
            return "dap2"
        return "dap2"

    def make_dataset(
        self,
    ):
        if self.protocol == "dap4":
            self.dataset_from_dap4()
        else:
            self.dataset_from_dap2()
        self.attach_das()

    def dataset_from_dap4(self):
        dmr_url = urllib.parse.urlunsplit(
            (self.scheme, self.netloc, self.path + ".dmr", self.query, self.fragment)
        )
        r = dapclient.net.GET(
            dmr_url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        dapclient.net.raise_for_status(r)
        dmr = safe_charset_text(r, self.user_charset)
        self.dataset = dmr_to_dataset(dmr)

    def dataset_from_dap2(self):
        dds_url = urllib.parse.urlunsplit(
            (self.scheme, self.netloc, self.path + ".dds", self.query, self.fragment)
        )
        r = dapclient.net.GET(
            dds_url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        dapclient.net.raise_for_status(r)
        dds = safe_charset_text(r, self.user_charset)
        self.dataset = dds_to_dataset(dds)

    def attach_das(self):
        # Also pull the DAS and add additional attributes
        das_url = urllib.parse.urlunsplit(
            (self.scheme, self.netloc, self.path + ".das", self.query, self.fragment)
        )
        r = dapclient.net.GET(
            das_url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        dapclient.net.raise_for_status(r)
        das = safe_charset_text(r, self.user_charset)
        add_attributes(self.dataset, parse_das(das))

    def add_proxies(self):
        if self.protocol == "dap4":
            self.add_dap4_proxies()
        else:
            self.add_dap2_proxies()

    def add_dap4_proxies(self):
        # remove any projection from the base_url, leaving selections
        for var in walk(self.dataset, dapclient.model.BaseType):
            var.data = BaseProxyDap4(
                self.base_url,
                var.name,
                var.dtype,
                var.shape,
                application=self.application,
                session=self.session,
                timeout=self.timeout,
            )
        for var in walk(self.dataset, dapclient.model.GridType):
            var.set_output_grid(self.output_grid)

    def add_dap2_proxies(self):
        # now add data proxies
        for var in walk(self.dataset, dapclient.model.BaseType):
            var.data = BaseProxyDap2(
                self.base_url,
                var.id,
                var.dtype,
                var.shape,
                application=self.application,
                session=self.session,
                timeout=self.timeout,
            )
        for var in walk(self.dataset, dapclient.model.SequenceType):
            template = copy.copy(var)
            var.data = SequenceProxy(
                self.base_url,
                template,
                application=self.application,
                session=self.session,
                timeout=self.timeout,
            )

        # apply projections
        for var in self.projection:
            target = self.dataset
            while var:
                token, index = var.pop(0)
                target = target[token]
                if isinstance(target, dapclient.model.BaseType):
                    target.data.slice = fix_slice(index, target.shape)
                elif isinstance(target, dapclient.model.GridType):
                    index = fix_slice(index, target.array.shape)
                    target.array.data.slice = index
                    for s, child in zip(index, target.maps):
                        target[child].data.slice = (s,)
                elif isinstance(target, dapclient.model.SequenceType):
                    target.data.slice = index

        # retrieve only main variable for grid types:
        for var in walk(self.dataset, dapclient.model.GridType):
            var.set_output_grid(self.output_grid)


def get_charset(r, user_charset):
    charset = r.charset
    if not charset:
        charset = user_charset
    return charset


def safe_charset_text(r, user_charset):
    if r.content_encoding == "gzip":
        return (
            gzip.GzipFile(fileobj=BytesIO(r.body))
            .read()
            .decode(get_charset(r, user_charset))
        )
    r.charset = get_charset(r, user_charset)
    return r.text


def safe_dds_and_data(r, user_charset):
    if r.content_encoding == "gzip":
        raw = gzip.GzipFile(fileobj=BytesIO(r.body)).read()
    else:
        raw = r.body
    dds, data = raw.split(b"\nData:\n", 1)
    return dds.decode(get_charset(r, user_charset)), data


def safe_dmr_and_data(r, user_charset):
    if r.content_encoding == "gzip":
        raw = gzip.GzipFile(fileobj=BytesIO(r.body)).read()
    else:
        raw = r.body
    logger.info("Saving and splitting dmr+")
    try:
        dmr, data = raw.split(b"</Dataset>", 1)
    except ValueError:
        logger.exception(f"Failed to split the following DMR+ \n {raw}")

        picked_response = str(codecs.encode(pickle.dumps(r), "base64").decode())
        logger.exception(
            "pickled response (base64): \n ----BEGIN PICKLE----- \n %s -----END PICKLE-----"
            % picked_response
        )

    dmr = dmr[4:] + b"</Dataset>"
    dmr = dmr.decode(get_charset(r, user_charset))
    data = data[3:]
    return dmr, data


class BaseProxyDap2:
    """A proxy for remote base types.

    This class behaves like a Numpy array, proxying the data from a base type
    on a remote dataset.
    """

    def __init__(
        self,
        baseurl,
        id,
        dtype,
        shape,
        slice_=None,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_charset="ascii",
    ):
        self.baseurl = baseurl
        self.id = id
        self.dtype = dtype
        self.shape = shape
        self.slice = slice_ or tuple(slice(None) for s in self.shape)
        self.application = application
        self.session = session
        self.timeout = timeout
        self.verify = verify
        self.user_charset = user_charset

    def __repr__(self):
        return "BaseProxy(%s)" % ", ".join(
            map(repr, [self.baseurl, self.id, self.dtype, self.shape, self.slice])
        )

    def __getitem__(self, index):
        # build download url

        index = combine_slices(self.slice, fix_slice(index, self.shape))
        scheme, netloc, path, query, fragment = urllib.parse.urlsplit(self.baseurl)
        url = urllib.parse.urlunsplit(
            (
                scheme,
                netloc,
                path + ".dods",
                urllib.parse.quote(self.id) + hyperslab(index) + "&" + query,
                fragment,
            )
        ).rstrip("&")

        # download and unpack data
        logger.info(f"Fetching URL: {url}")
        r = dapclient.net.GET(
            url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )

        dapclient.net.raise_for_status(r)
        dds, data = safe_dds_and_data(r, self.user_charset)

        # Parse received dataset:
        dataset = dds_to_dataset(dds)
        dataset.data = unpack_dap2_data(BytesReader(data), dataset)
        return dataset[self.id].data

    def __len__(self):
        return self.shape[0]

    def __iter__(self):
        return iter(self[:])

    # Comparisons return a boolean array
    def __eq__(self, other):
        return self[:] == other

    def __ne__(self, other):
        return self[:] != other

    def __ge__(self, other):
        return self[:] >= other

    def __le__(self, other):
        return self[:] <= other

    def __gt__(self, other):
        return self[:] > other

    def __lt__(self, other):
        return self[:] < other


class BaseProxyDap4(BaseProxyDap2):
    def __init__(
        self,
        baseurl,
        id,
        dtype,
        shape,
        slice_=None,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_charset="ascii",
    ):
        self.baseurl = baseurl
        self.id = id
        self.dtype = dtype
        self.shape = shape
        self.slice = slice_ or tuple(slice(None) for s in self.shape)
        self.application = application
        self.session = session
        self.timeout = timeout
        self.verify = verify
        self.user_charset = user_charset

    def __repr__(self):
        return "Dap4BaseProxy(%s)" % ", ".join(
            map(repr, [self.baseurl, self.id, self.dtype, self.shape, self.slice])
        )

    def __getitem__(self, index):
        # build download url
        index = combine_slices(self.slice, fix_slice(index, self.shape))
        scheme, netloc, path, query, fragment = urllib.parse.urlsplit(self.baseurl)
        ce = "dap4.ce=" + urllib.parse.quote(self.id) + hyperslab(index) + query
        url = urllib.parse.urlunsplit(
            (scheme, netloc, path + ".dap", ce, fragment)
        ).rstrip("&")

        # download and unpack data
        logger.info(f"Fetching URL: {url}")

        r = dapclient.net.GET(
            url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )

        dapclient.net.raise_for_status(r)
        dmr, data = safe_dmr_and_data(r, self.user_charset)

        # Parse received dataset:
        dataset = dmr_to_dataset(dmr)
        dataset = unpack_dap4_data(BytesReader(data), dataset)

        self.checksum = dataset[self.id].attributes["checksum"]
        self.data = dataset[self.id].data
        return self.data


class SequenceProxy:
    """A proxy for remote sequences.

    This class behaves like a Numpy structured array, proxying the data from a
    sequence on a remote dataset. The data is streamed from the dataset,
    meaning it can be treated one record at a time before the whole data is
    downloaded.

    """

    shape = ()

    def __init__(
        self,
        baseurl,
        template,
        selection=None,
        slice_=None,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
    ):
        self.baseurl = baseurl
        self.template = template
        self.selection = selection or []
        self.slice = slice_ or (slice(None),)
        self.application = application
        self.session = session
        self.timeout = timeout
        self.verify = verify

        # this variable is true when only a subset of the children are selected
        self.sub_children = False

    @property
    def dtype(self):
        return self.template.dtype

    def __repr__(self):
        return "SequenceProxy(%s)" % ", ".join(
            map(repr, [self.baseurl, self.template, self.selection, self.slice])
        )

    def __copy__(self):
        """Return a lightweight copy of the object."""
        return self.__class__(
            self.baseurl,
            self.template,
            self.selection[:],
            self.slice[:],
            self.application,
        )

    def __getitem__(self, key):
        """Return a new object representing a subset of the data."""
        out = copy.copy(self)

        # return the data for a children
        if isinstance(key, str):
            out.template = out.template[key]

        # return a new object with requested columns
        elif isinstance(key, list):
            out.sub_children = True
            out.template._visible_keys = key

        # return a copy with the added constraints
        elif isinstance(key, dapclient.handlers.lib.ConstraintExpression):
            out.selection.extend(str(key).split("&"))

        # slice data
        else:
            if isinstance(key, int):
                key = slice(key, key + 1)
            out.slice = combine_slices(self.slice, (key,))

        return out

    @property
    def url(self):
        """Return url from where data is fetched."""
        scheme, netloc, path, query, fragment = urllib.parse.urlsplit(self.baseurl)
        url = urllib.parse.urlunsplit(
            (
                scheme,
                netloc,
                path + ".dods",
                self.id + hyperslab(self.slice) + "&" + "&".join(self.selection),
                fragment,
            )
        ).rstrip("&")

        return url

    @property
    def id(self):
        """Return the id of this sequence."""
        if self.sub_children:
            id_ = ",".join(
                urllib.parse.quote(child.id) for child in self.template.children()
            )
        else:
            id_ = urllib.parse.quote(self.template.id)
        return id_

    def __iter__(self):
        # download and unpack data
        r = dapclient.net.GET(
            self.url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        dapclient.net.raise_for_status(r)

        i = r.app_iter
        if not hasattr(i, "__next__"):
            i = iter(i)

        # Fast forward past the DDS header
        # the pattern could span chunk boundaries though so make sure to check
        pattern = b"Data:\n"
        last_chunk = find_pattern_in_string_iter(pattern, i)

        if last_chunk is None:
            raise ValueError(f"Could not find data segment in response from {self.url}")

        # Then construct a stream consisting of everything from
        # 'Data:\n' to the end of the chunk + the rest of the stream
        def stream_start():
            yield last_chunk

        stream = StreamReader(chain(stream_start(), i))

        return unpack_sequence(stream, self.template)

    def __eq__(self, other):
        return dapclient.handlers.lib.ConstraintExpression(f"{self.id}={encode(other)}")

    def __ne__(self, other):
        return dapclient.handlers.lib.ConstraintExpression(
            f"{self.id}!={encode(other)}"
        )

    def __ge__(self, other):
        return dapclient.handlers.lib.ConstraintExpression(
            f"{self.id}>={encode(other)}"
        )

    def __le__(self, other):
        return dapclient.handlers.lib.ConstraintExpression(
            f"{self.id}<={encode(other)}"
        )

    def __gt__(self, other):
        return dapclient.handlers.lib.ConstraintExpression(f"{self.id}>{encode(other)}")

    def __lt__(self, other):
        return dapclient.handlers.lib.ConstraintExpression(f"{self.id}<{encode(other)}")


def unpack_sequence(stream, template):
    """Unpack data from a sequence, yielding records."""
    # is this a sequence or a base type?
    sequence = isinstance(template, dapclient.model.SequenceType)

    # if there are no children, we use the template as the only column
    cols = list(template.children()) or [template]

    # if there are no strings and no nested sequences we can unpack record by
    # record easily
    simple = all(
        isinstance(c, dapclient.model.BaseType) and c.dtype.char not in "SU"
        for c in cols
    )

    if simple:
        dtype = numpy.dtype([("", c.dtype, c.shape) for c in cols])
        marker = stream.read(4)
        while marker == START_OF_SEQUENCE:
            rec = numpy.frombuffer(stream.read(dtype.itemsize), dtype=dtype)[0]
            if not sequence:
                rec = rec[0]
            yield rec
            marker = stream.read(4)
    else:
        marker = stream.read(4)
        while marker == START_OF_SEQUENCE:
            rec = unpack_children(stream, template)
            if not sequence:
                rec = rec[0]
            else:
                rec = tuple(rec)
            yield rec
            marker = stream.read(4)


def unpack_children(stream, template):
    """Unpack children from a structure, returning their data."""
    cols = list(template.children()) or [template]

    out = []
    for col in cols:
        # sequences and other structures
        if isinstance(col, dapclient.model.SequenceType):
            out.append(
                dapclient.handlers.lib.IterData(list(unpack_sequence(stream, col)), col)
            )
        elif isinstance(col, dapclient.model.StructureType):
            out.append(tuple(unpack_children(stream, col)))

        # unpack arrays
        else:
            out.extend(convert_stream_to_list(stream, col.dtype, col.shape, col.id))
    return out


def convert_stream_to_list(stream, parser_dtype, shape, id):
    out = []
    response_dtype = DAP2_response_dtypemap(parser_dtype)
    if shape:
        n = numpy.frombuffer(stream.read(4), DAP2_ARRAY_LENGTH_NUMPY_TYPE)[0]
        count = response_dtype.itemsize * n
        if response_dtype.char in "S":
            # Consider on 'S' and not 'SU' because
            # response_dtype.char should never be
            data = []
            for _ in range(n):
                k = numpy.frombuffer(stream.read(4), DAP2_ARRAY_LENGTH_NUMPY_TYPE)[0]
                data.append(stream.read(k))
                stream.read(-k % 4)
            out.append(
                numpy.array([str(x.decode("ascii")) for x in data], "S").reshape(shape)
            )
        else:
            stream.read(4)  # read additional length
            try:
                out.append(
                    numpy.frombuffer(stream.read(count), response_dtype)
                    .astype(parser_dtype)
                    .reshape(shape)
                )
            except ValueError as exc:
                if str(exc) == "total size of new array must be unchanged":
                    # server-side failure.
                    # it is expected that the user should be mindful of this:
                    raise RuntimeError(
                        """
                            Variable {urllib.parse.quote(id)} could not be
                            properly retrieved. To avoid this error consider
                            using open_url(..., output_grid=False).
                            """
                    ) from exc
                raise
            if response_dtype.char == "B":
                # Unsigned Byte type is packed to multiples of 4 bytes:
                stream.read(-n % 4)

    # special types: strings and bytes
    elif response_dtype.char in "S":
        # Consider on 'S' and not 'SU' because
        # response_dtype.char should never be
        # 'U'
        k = numpy.frombuffer(stream.read(4), DAP2_ARRAY_LENGTH_NUMPY_TYPE)[0]
        out.append(str(stream.read(k).decode("ascii")))
        stream.read(-k % 4)
    # usual data
    else:
        out.append(
            numpy.frombuffer(
                stream.read(response_dtype.itemsize), response_dtype
            ).astype(parser_dtype)[0]
        )
        if response_dtype.char == "B":
            # Unsigned Byte type is packed to multiples of 4 bytes:
            stream.read(3)
    return out


def unpack_dap2_data(xdr_stream, dataset):
    """Unpack a string of encoded data, returning data as lists."""
    return unpack_children(xdr_stream, dataset)


def decode_chunktype(chunk_type):
    encoding = f"{chunk_type:03b}"
    if sys.byteorder == "little":
        # If our machine's byteorder is little, we need to swap since the chunk_type is always big endian
        encoding = encoding[::-1]
    last_chunk = bool(int(encoding[0]))
    error = bool(int(encoding[1]))
    endian = {"0": ">", "1": "<"}[encoding[2]]
    return last_chunk, error, endian


def get_count(variable):
    count = int(numpy.array(variable.shape).prod())
    item_size = numpy.dtype(variable.dtype).itemsize
    count = count * item_size
    return count


def decode_variable(buffer, start, stop, variable, endian):
    dtype = variable.dtype
    dtype = dtype.newbyteorder(endian)
    data = numpy.frombuffer(buffer[start:stop], dtype=dtype).astype(dtype)
    data = data.reshape(variable.shape)
    return data


def stream2bytearray(xdr_stream):
    last = False
    buffer = bytearray()
    while not last:
        chunk = numpy.frombuffer(xdr_stream.read(4), dtype=">u4")
        chunk_size = (chunk & 0x00FFFFFF)[0]
        chunk_type = ((chunk >> 24) & 0xFF)[0]
        last, error, endian = decode_chunktype(chunk_type)
        buffer.extend(xdr_stream.read(chunk_size))
    return buffer


def get_endianness(xdr_stream):
    chunk_header = xdr_stream.peek(4)[0:4]
    chunk_header = numpy.frombuffer(chunk_header, dtype=">u4")[0]
    chunk_type = (chunk_header >> 24) & 0xFF
    last, error, endian = decode_chunktype(chunk_type)
    return endian


def unpack_dap4_data(xdr_stream, dataset):
    endian = get_endianness(xdr_stream)
    checksum_dtype = numpy.dtype(endian + "u4")
    buffer = stream2bytearray(xdr_stream)

    start = 0
    for variable_name in dataset:
        variable = dataset[variable_name]
        count = get_count(variable)
        stop = start + count
        data = decode_variable(
            buffer, start=start, stop=stop, variable=variable, endian=endian
        )
        checksum = numpy.frombuffer(
            buffer[stop : stop + 4], dtype=checksum_dtype
        ).byteswap("=")
        if isinstance(variable, dapclient.model.BaseType):
            variable._set_data(data)
        elif isinstance(variable, dapclient.model.GridType):
            variable._set_data([data.data])
        variable.attributes["checksum"] = checksum
        # Jump over the 4 byte chunk_header
        start = stop + 4
    return dataset


def find_pattern_in_string_iter(pattern, i):
    last_chunk = b""
    length = len(pattern)
    for this_chunk in i:
        last_chunk += this_chunk
        m = re.search(pattern, last_chunk)
        if m:
            return last_chunk[m.end() :]
        last_chunk = last_chunk[-length:]


def dump():  # pragma: no cover
    """Unpack dods response into lists.

    Return pretty-printed data.

    """
    dods = sys.stdin.read()
    dds, xdrdata = dods.split(b"\nData:\n", 1)
    dataset = dds_to_dataset(dds)
    xdr_stream = io.BytesIO(xdrdata)
    data = unpack_dap2_data(xdr_stream, dataset)
    pprint.pprint(data)



"""Datasets for testing.

This module defines a few datasets for testing, covering the whole DAP data
model.

"""
import os
from collections import OrderedDict

import numpy as np

from dapclient.client import open_file
from dapclient.handlers.lib import IterData
from dapclient.model import BaseType, DatasetType, GridType, SequenceType, StructureType

# Note that DAP2 does not support signed bytes (signed 8bits integers).

# A very simple sequence: flat and with no strings. This sequence can be mapped
# directly to a Numpy structured array, and can be easily encoded and decoded
# in the DAP spec.
VerySimpleSequence = DatasetType("VerySimpleSequence")
VerySimpleSequence["sequence"] = SequenceType("sequence")
VerySimpleSequence["sequence"]["byte"] = BaseType("byte")
VerySimpleSequence["sequence"]["int"] = BaseType("int")
VerySimpleSequence["sequence"]["float"] = BaseType("float")
VerySimpleSequence["sequence"].data = np.array(
    [
        (0, 1, 10.0),
        (1, 2, 20.0),
        (2, 3, 30.0),
        (3, 4, 40.0),
        (4, 5, 50.0),
        (5, 6, 60.0),
        (6, 7, 70.0),
        (7, 8, 80.0),
    ],
    dtype=[("byte", np.ubyte), ("int", "i4"), ("float", "f4")],
)


# A nested sequence.
NestedSequence = DatasetType("NestedSequence")
NestedSequence["location"] = SequenceType("location")
NestedSequence["location"]["lat"] = BaseType("lat")
NestedSequence["location"]["lon"] = BaseType("lon")
NestedSequence["location"]["elev"] = BaseType("elev")
NestedSequence["location"]["time_series"] = SequenceType("time_series")
NestedSequence["location"]["time_series"]["time"] = BaseType("time")
NestedSequence["location"]["time_series"]["slp"] = BaseType("slp")
NestedSequence["location"]["time_series"]["wind"] = BaseType("wind")
NestedSequence["location"].data = IterData(
    [
        (1, 1, 1, [(10, 11, 12), (21, 22, 23)]),
        (2, 4, 4, [(15, 16, 17)]),
        (3, 6, 9, []),
        (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
    ],
    NestedSequence.location,
)


# A simple array with bytes, strings and shorts. These types require special
# encoding for the DODS response.
SimpleArray = DatasetType("SimpleArray")
SimpleArray["byte"] = BaseType("byte", np.arange(5, dtype=np.ubyte))
SimpleArray["string"] = BaseType("string", np.array(["one", "two"]))
SimpleArray["short"] = BaseType("short", np.array(1, dtype="h"))


DODS = os.path.join(os.path.dirname(__file__), "data/rainfall_time_malaysia.cdp.dods")
DAS = os.path.join(os.path.dirname(__file__), "data/rainfall_time_malaysia.cdp.das")
dapper = open_file(DODS, DAS)


# dataset from http://test.opendap.org:8080/dods/dts/D1.asc
D1 = DatasetType("EOSDB.DBO", type="Drifters")
D1["Drifters"] = SequenceType("Drifters")
D1["Drifters"]["instrument_id"] = BaseType("instrument_id")
D1["Drifters"]["location"] = BaseType("location")
D1["Drifters"]["latitude"] = BaseType("latitude")
D1["Drifters"]["longitude"] = BaseType("longitude")
D1.Drifters.data = np.array(
    np.rec.fromrecords(
        list(
            zip(
                [
                    "This is a data test string (pass {}).".format(1 + i * 2)
                    for i in range(5)
                ],
                [
                    "This is a data test string (pass {}).".format(i * 2)
                    for i in range(5)
                ],
                [1000.0, 999.95, 999.80, 999.55, 999.20],
                [999.95, 999.55, 998.75, 997.55, 995.95],
            )
        ),
        names=list(D1.Drifters.keys()),
    )
)


# testing structures
SimpleStructure = DatasetType("SimpleStructure")
SimpleStructure["types"] = StructureType(
    name="types",
    key="value",
    nested=OrderedDict(
        [
            ("string", "bar"),
            ("list", [42, 43]),
            ("array", np.array(1)),
            ("float", 1000.0),
        ]
    ),
)
SimpleStructure["types"]["b"] = BaseType("b", np.array(-10, np.byte))
SimpleStructure["types"]["ub"] = BaseType("ub", np.array(10, np.ubyte))
SimpleStructure["types"]["i32"] = BaseType("i32", np.array(-10, np.int32))
SimpleStructure["types"]["ui32"] = BaseType("ui32", np.array(10, np.uint32))
SimpleStructure["types"]["i16"] = BaseType("i16", np.array(-10, np.int16))
SimpleStructure["types"]["ui16"] = BaseType("ui16", np.array(10, np.uint16))
SimpleStructure["types"]["f32"] = BaseType("f32", np.array(100.0, np.float32))
SimpleStructure["types"]["f64"] = BaseType("f64", np.array(1000.0, np.float64))
SimpleStructure["types"]["s"] = BaseType(
    "s", np.array("This is a data test string (pass 0).")
)
SimpleStructure["types"]["u"] = BaseType("u", np.array("http://www.dods.org"))
SimpleStructure["types"]["U"] = BaseType("U", np.array("test unicode", str))


# test grid
rain = DatasetType("test")
rain["rain"] = GridType("rain")
rain["rain"]["rain"] = BaseType(
    "rain", np.arange(6).reshape(2, 3), dimensions=("y", "x")
)
rain["rain"]["x"] = BaseType("x", np.arange(3), units="degrees_east")
rain["rain"]["y"] = BaseType("y", np.arange(2), units="degrees_north")


# test for ``bounds`` function
bounds = DatasetType("test")
bounds["sequence"] = SequenceType("sequence")
bounds["sequence"]["lon"] = BaseType("lon", axis="X")
bounds["sequence"]["lat"] = BaseType("lat", axis="Y")
bounds["sequence"]["depth"] = BaseType("depth", axis="Z")
bounds["sequence"]["time"] = BaseType("time", axis="T", units="days since 1970-01-01")
bounds["sequence"]["measurement"] = BaseType("measurement")
bounds.sequence.data = np.array(
    np.rec.fromrecords(
        [
            (100, -10, 0, -1, 42),
            (200, 10, 500, 1, 43),
        ],
        names=list(bounds.sequence.keys()),
    )
)


# test for density
ctd = DatasetType("ctd")
ctd["cast"] = SequenceType("cast")
ctd["cast"]["temperature"] = BaseType("temperature")
ctd["cast"]["salinity"] = BaseType("salinity")
ctd["cast"]["pressure"] = BaseType("pressure")
ctd.cast.data = np.array(
    np.rec.fromrecords(
        [
            (21, 35, 0),
            (15, 35, 100),
        ],
        names=list(ctd.cast.keys()),
    )
)

# a simple sequence, simulating a CTD profile
SimpleSequence = DatasetType(
    "SimpleSequence", description="A simple sequence for testing.", nested={"value": 42}
)
SimpleSequence["cast"] = SequenceType("cast")
SimpleSequence["cast"]["id"] = BaseType("id")
SimpleSequence["cast"]["lon"] = BaseType("lon", axis="X")
SimpleSequence["cast"]["lat"] = BaseType("lat", axis="Y")
SimpleSequence["cast"]["depth"] = BaseType("depth", axis="Z")
SimpleSequence["cast"]["time"] = BaseType(
    "time", axis="T", units="days since 1970-01-01"
)
SimpleSequence["cast"]["temperature"] = BaseType("temperature")
SimpleSequence["cast"]["salinity"] = BaseType("salinity")
SimpleSequence["cast"]["pressure"] = BaseType("pressure")
SimpleSequence["cast"].data = np.array(
    np.rec.fromrecords(
        [
            ("1", 100, -10, 0, -1, 21, 35, 0),
            ("2", 200, 10, 500, 1, 15, 35, 100),
        ],
        names=list(SimpleSequence["cast"].keys()),
    )
)
# SimpleSequence["cast"].data = IterData([
#     ("1", 100, -10,   0, -1, 21, 35,   0),
#     ("2", 200,  10, 500,  1, 15, 35, 100),
# ], SimpleSequence.cast)

# a simple grid
SimpleGrid = DatasetType("SimpleGrid", description="A simple grid for testing.")
SimpleGrid["SimpleGrid"] = GridType("SimpleGrid")
SimpleGrid["SimpleGrid"]["SimpleGrid"] = BaseType(
    "SimpleGrid", np.arange(6).reshape(2, 3), dimensions=("y", "x")
)
SimpleGrid["x"] = SimpleGrid["SimpleGrid"]["x"] = BaseType(
    "x", np.arange(3), axis="X", units="degrees_east"
)
SimpleGrid["y"] = SimpleGrid["SimpleGrid"]["y"] = BaseType(
    "y", np.arange(2), axis="Y", units="degrees_north"
)


# a faulty grid
FaultyGrid = DatasetType("FaultyGrid", description="A faulty grid for testing.")
FaultyGrid["FaultyGrid"] = GridType("FaultyGrid")
FaultyGrid["FaultyGrid"]["FaultyGrid"] = BaseType(
    "FaultyGrid", np.arange(6).reshape(2, 3), dimensions=("y", "x")
)
FaultyGrid["x"] = FaultyGrid["FaultyGrid"]["x"] = BaseType(
    "x", np.arange(3), axis="X", code=1
)
FaultyGrid["y"] = FaultyGrid["FaultyGrid"]["y"] = BaseType(
    "y", np.arange(2), axis="Y", code=np.int32([])
)



"""A DMR parser."""

import collections
import copy
import re
from xml.etree import ElementTree as ET

import numpy as np

import dapclient.lib
import dapclient.model

constructors = ("grid", "sequence", "structure")
name_regexp = r'[\w%!~"\'\*-]+'
dmr_atomic_types = (
    "Int8",
    "UInt8",
    "Byte",
    "Char",
    "Int16",
    "UInt16",
    "Int32",
    "UInt32",
    "Int64",
    "UInt64",
    "Float32",
    "Float64",
)

namespace = {"": "http://xml.opendap.org/ns/DAP/4.0#"}


def dap4_to_numpy_typemap(type_string):
    """
    This function takes a numpy dtype object
    and returns a dtype object that is compatible with
    the DAP2 specification.
    """
    dtype_str = dapclient.lib.DAP4_TO_NUMPY_PARSER_TYPEMAP[type_string]
    return np.dtype(dtype_str)


def get_variables(node, prefix=""):
    variables = collections.OrderedDict()
    group_name = node.get("name")
    if group_name is None:
        return variables
    if node.tag != "Dataset":
        prefix = prefix + "/" + group_name
    for subnode in node:
        if subnode.tag in dmr_atomic_types:
            name = subnode.get("name")
            if prefix != "":
                name = prefix + "/" + name
            variables[name] = {"element": subnode}
        variables.update(get_variables(subnode, prefix))
    return variables


def get_named_dimensions(node, prefix=""):
    dimensions = {}
    group_name = node.get("name")
    if group_name is None:
        return dimensions
    if node.tag != "Dataset":
        prefix = prefix + "/" + group_name
    for subnode in node:
        if subnode.tag == "Dimension":
            name = subnode.get("name")
            if prefix != "":
                name = prefix + "/" + name
            dimensions[name] = int(subnode.attrib["size"])
        dimensions.update(get_named_dimensions(subnode, prefix))
    return dimensions


def get_dtype(element):
    dtype = element.tag
    dtype = dap4_to_numpy_typemap(dtype)
    return dtype


def get_attributes(element):
    attributes = {}
    attribute_elements = element.findall("Attribute")
    for attribute_element in attribute_elements:
        name = attribute_element.get("name")
        value = attribute_element.find("Value").text
        attributes[name] = value
    return attributes


def get_dim_names(element):
    # Not to be confused with dimensions
    dimension_elements = element.findall("Dim")
    dimensions = []
    for dimension_element in dimension_elements:
        name = dimension_element.get("name")
        if name is None:
            # We might have unnamed dimensions
            return dimensions
        if name.find("/", 1) == -1:
            # If this is a root Dimension, we remove the leading slash
            name = name.replace("/", "")
        dimensions.append(name)
    return dimensions


def get_dim_sizes(element):
    dimension_elements = element.findall("Dim")
    dimension_sizes = ()
    for dimension_element in dimension_elements:
        name = dimension_element.get("name")
        if name is None:
            size = int(dimension_element.get("size"))
            dimension_sizes += (size,)
    return dimension_sizes


def has_map(element):
    maps = element.findall("Map")
    if len(maps) > 0:
        return True
    return False


def dmr_to_dataset(dmr):
    """Return a dataset object from a DMR representation."""

    # Parse the DMR. First dropping the namespace
    dmr = re.sub(' xmlns="[^"]+"', "", dmr, count=1)
    dom_et = ET.fromstring(dmr)

    variables = get_variables(dom_et)
    named_dimensions = get_named_dimensions(dom_et)

    # Add size entry for dimension variables
    for name, size in named_dimensions.items():
        if name in variables:
            variables[name]["size"] = size

    # Bootstrap variables
    for name, variable in variables.items():
        variable["name"] = name
        variable["attributes"] = get_attributes(variable["element"])
        variable["dtype"] = get_dtype(variable["element"])
        variable["dims"] = get_dim_names(variable["element"])
        variable["has_map"] = has_map(variable["element"])
        variable["shape"] = get_dim_sizes(variable["element"])

    # Add size entry for dimension variables
    for name, size in named_dimensions.items():
        if name not in variables:
            # We might have dimensions that only have a declaration, so we need
            # to add them to the variables
            variables[name] = {
                "name": name,
                "size": size,
                "dims": [name],
                "dtype": "int",
                "has_map": False,
                "attributes": {},
                "shape": (),
            }

    # Add shape element to variables
    for name, variable in variables.items():
        for dim in variable["dims"]:
            variable["shape"] += (variables[dim]["size"],)

    # Convert the ordered dictionary to dataset
    dataset_name = dom_et.attrib["name"]
    dataset = dapclient.model.DatasetType(dataset_name)
    for name, variable in variables.items():
        data = DummyData(dtype=variable["dtype"], shape=variable["shape"])
        array = dapclient.model.BaseType(
            name=variable["name"], data=data, dimensions=variable["dims"]
        )
        if variable["has_map"]:
            var = dapclient.model.GridType(name=variable["name"])
            var[name] = array
            for dim in variable["dims"]:
                var[dim] = copy.copy(dataset[dim])
        else:
            var = array
        var.attributes = variable["attributes"]
        dataset[var.name] = var

    return dataset


class DMRParser:
    """A parser for the DMR."""

    def __init__(self, dmr):
        self.dmr = dmr


class DummyData:
    def __init__(self, dtype, shape):
        self.dtype = dtype
        self.shape = shape



import pytest

import dapclient.client


def test_coads():
    url = "dap4://test.opendap.org/opendap/hyrax/data/nc/coads_climatology.nc"
    pydap_ds = dapclient.client.open_url(url)
    data = pydap_ds["COADSX"][10:12:1]
    assert data is not None


@pytest.mark.skip(reason="takes a really long time")
def test_groups():
    url = "dap4://test.opendap.org:8080/opendap/dmrpp_test_files/ATL03_20181228015957_13810110_003_01.2var.h5.dmrpp"
    pydap_ds = dapclient.client.open_url(url)
    data = pydap_ds["/gt1r/bckgrd_atlas/bckgrd_int_height"][0:10]
    assert data is not None


@pytest.mark.skip(reason="bug in testserver")
def test_maps():
    url = "dap4://test.opendap.org:8080/opendap/hyrax/data/nc/coads_climatology.nc"
    pydap_ds = dapclient.client.open_url(url)
    data = pydap_ds["SST"][0:2:1, 40:42:1, 1:10:1]
    assert data is not None


if __name__ == "__main__":
    test_maps()



"""Test DAP4 parsing functions."""

import os

import dapclient.parsers.dmr


def load_dmr_file(file_path):
    abs_path = os.path.join(os.path.dirname(__file__), file_path)
    with open(abs_path) as dmr_file:
        dmr = dmr_file.read()
    dataset = dapclient.parsers.dmr.dmr_to_dataset(dmr)
    return dataset


def test_single_scalar():
    """Test a single scalar case."""
    single_scalar_dmr = """<Dataset name="foo"><Int32 name="x"/></Dataset>"""
    dataset = dapclient.parsers.dmr.dmr_to_dataset(single_scalar_dmr)
    assert dataset["x"].dtype == ">i4"


def test_coads_climatology2():
    dataset = load_dmr_file("data/dmrs/coads_climatology.nc.dmr")
    assert dataset["SST"].attributes["long_name"] == "SEA SURFACE TEMPERATURE"
    assert dataset["SST"].attributes["missing_value"] == "-9.99999979e+33"
    assert dataset["AIRT"].shape == (12, 90, 180)
    assert dataset["VWND"].dtype.str == ">f4"


def test_atl03():
    # Contains groups
    dataset = load_dmr_file(
        "data/dmrs/ATL03_20181228015957_13810110_003_01.2var.h5.dmrpp.dmr"
    )
    assert (
        dataset["/gt1r/bckgrd_atlas/bckgrd_int_height"].attributes["contentType"]
        == "modelResult"
    )


def test_jpl():
    # Contains groups
    dataset = load_dmr_file(
        "data/dmrs/20220102090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1.dmr"
    )
    assert dataset["sea_ice_fraction"].dimensions == ["time", "lat", "lon"]


def test_mod05():
    dataset = load_dmr_file(
        "data/dmrs/MOD05_L2.A2019336.2315.061.2019337071952.hdf.dmr"
    )
    assert dataset["Water_Vapor_Infrared"].dimensions == [
        "Cell_Along_Swath_5km",
        "Cell_Across_Swath_5km",
    ]



"""The ASCII response.

The ASCII response is an unofficial response used to return the data as ASCII.
dapclient's implementation is reverse engineered from the official server.

"""

import copy
from functools import singledispatch

import numpy as np
from pkg_resources import get_distribution

from dapclient.lib import encode
from dapclient.model import BaseType, SequenceType, StructureType
from dapclient.responses.dds import dds
from dapclient.responses.lib import BaseResponse


class ASCIIResponse(BaseResponse):
    """The ASCII response."""

    __version__ = get_distribution("dapclient").version

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_ascii"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        for line in dds(self.dataset):
            yield line.encode("ascii")
        yield (45 * "-" + "\n").encode("ascii")

        for line in ascii(self.dataset):
            yield line.encode("ascii")


@singledispatch
def ascii(var, printname=True):
    """A single dispatcher for the ASCII response."""
    raise StopIteration


@ascii.register(SequenceType)
def _sequenctype(var, printname=True):
    yield ", ".join([child.id for child in var.children()])
    yield "\n"
    for rec in var.iterdata():
        out = copy.copy(var)
        out.__class__ = StructureType
        out.data = rec
        for i, line in enumerate(ascii(out, printname=False)):
            line = line.strip()
            if line and i > 0:
                yield ", "
            yield line
        yield "\n"


@ascii.register(StructureType)
def _structuretype(var, printname=True):
    for child in var.children():
        yield from ascii(child, printname)
        yield "\n"


@ascii.register(BaseType)
def _basetype(var, printname=True):
    if printname:
        yield var.id
        yield "\n"

    if not getattr(var, "shape", ()):
        yield encode(var.data)
    else:
        for indexes, value in zip(np.ndindex(var.shape), var.data.flat):
            yield "{indexes} {value}\n".format(
                indexes="[" + "][".join([str(idx) for idx in indexes]) + "]",
                value=encode(value),
            )



"""The DDS response.

The DDS response builds a representation of the structure of the dataset,
informing which variables are contained, their shape, type and dimensions.
Together with the DAS, the DDS describes all metadata associated with a given
dataset, allowing clients to introspect the variables and request data as
necessary.

"""

from functools import singledispatch

from pkg_resources import get_distribution

from dapclient.lib import NUMPY_TO_DAP2_TYPEMAP
from dapclient.model import BaseType, DatasetType, GridType, SequenceType, StructureType
from dapclient.responses.lib import BaseResponse

INDENT = " " * 4


class DDSResponse(BaseResponse):
    """The DDS response."""

    __version__ = get_distribution("dapclient").version

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_dds"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        for line in dds(self.dataset):
            yield line.encode("ascii")


@singledispatch
def dds(var, level=0, sequence=0):
    """Single dispatcher that generates the DDS response."""
    raise StopIteration


@dds.register(DatasetType)
def _(var, level=0, sequence=0):
    yield f"{level * INDENT}Dataset {{\n"
    for child in var.children():
        yield from dds(child, level + 1, sequence)
    yield f"{level * INDENT}}} {var.name};\n"


@dds.register(SequenceType)
def _sequencetype(var, level=0, sequence=0):
    yield f"{level * INDENT}Sequence {{\n"
    for child in var.children():
        yield from dds(child, level + 1, sequence + 1)
    yield f"{level * INDENT}}} {var.name};\n"


@dds.register(StructureType)
def _structuretype(var, level=0, sequence=0):
    yield f"{level * INDENT}Structure {{\n"
    for child in var.children():
        yield from dds(child, level + 1, sequence)
    yield f"{level * INDENT}}} {var.name};\n"


@dds.register(GridType)
def _gridtype(var, level=0, sequence=0):
    yield f"{level * INDENT}Grid {{\n"

    yield f"{(level + 1) * INDENT}Array:\n"
    yield from dds(var.array, level + 2, sequence)

    yield f"{(level + 1) * INDENT}Maps:\n"
    for map_ in var.maps.values():
        yield from dds(map_, level + 2, sequence)

    yield f"{level * INDENT}}} {var.name};\n"


@dds.register(BaseType)
def _basetype(var, level=0, sequence=0):
    shape = var.shape[sequence:]

    if var.dimensions:
        shape = "".join(map("[{0[0]} = {0[1]}]".format, zip(var.dimensions, shape)))
    elif len(shape) == 1:
        shape = f"[{var.name} = {shape[0]}]"
    else:
        shape = "".join(f"[{len}]" for len in shape)

    yield "{indent}{type} {name}{shape};\n".format(
        indent=level * INDENT,
        type=NUMPY_TO_DAP2_TYPEMAP[var.dtype.char],
        name=var.name,
        shape=shape,
    )



"""DAP exceptions.

These exceptions are mostly used by the server. When an exception is captured,
proper error message is displayed (according to the DAP 2.0 spec), with
information about the exception.

"""


class DapError(Exception):
    """Base DAP exception."""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ClientError(DapError):
    """Generic error with the client."""


class ServerError(DapError):
    """Generic error with the server."""


class ConstraintExpressionError(ServerError):
    """Exception raised when an invalid constraint expression is given."""


class HandlerError(DapError):
    """Generic error with a handler."""


class ExtensionNotSupportedError(HandlerError):
    """Exception raised when opening a file not supported by any handlers."""


class OpenFileError(HandlerError):
    """Exception raised when unable to open a file."""



"""The DAS response.

The DAS response describes the attributes associated with a dataset and its
variables. Together with the DDS the DAS response completely describes the
metadata of a dataset, allowing it to be introspected and data to be
downloaded.
"""

from collections.abc import Iterable
from functools import singledispatch

import numpy as np
from pkg_resources import get_distribution

from dapclient.lib import NUMPY_TO_DAP2_TYPEMAP, encode, quote
from dapclient.model import BaseType, DatasetType, GridType, SequenceType, StructureType
from dapclient.responses.lib import BaseResponse

INDENT = " " * 4


class DASResponse(BaseResponse):
    """The DAS response."""

    __version__ = get_distribution("dapclient").version

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_das"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        for line in das(self.dataset):
            try:
                yield line.encode("ascii")
            except UnicodeDecodeError:
                yield line.encode("UTF-8")


@singledispatch
def das(var, level=0):
    """Single dispatcher that generates the DAS response."""
    raise StopIteration


@das.register(DatasetType)
def _datasettype(var, level=0):
    yield f"{level * INDENT}Attributes {{\n"

    for attr in sorted(var.attributes.keys()):
        values = var.attributes[attr]
        yield from build_attributes(attr, values, level + 1)

    for child in var.children():
        yield from das(child, level=level + 1)
    yield f"{level * INDENT}}}\n"


@das.register(StructureType)
@das.register(SequenceType)
def _structuretype(var, level=0):
    yield f"{level * INDENT}{var.name} {{\n"

    for attr in sorted(var.attributes.keys()):
        values = var.attributes[attr]
        yield from build_attributes(attr, values, level + 1)

    for child in var.children():
        yield from das(child, level=level + 1)
    yield f"{level * INDENT}}}\n"


@das.register(BaseType)
@das.register(GridType)
def _basetypegridtype(var, level=0):
    yield f"{level * INDENT}{var.name} {{\n"

    for attr in sorted(var.attributes.keys()):
        values = var.attributes[attr]
        if np.asarray(values).size > 0:
            yield from build_attributes(attr, values, level + 1)
    yield f"{level * INDENT}}}\n"


def build_attributes(attr, values, level=0):
    """Recursive function to build the DAS."""
    # check for metadata
    if isinstance(values, dict):
        yield f"{level * INDENT}{attr} {{\n"
        for k, v in values.items():
            yield from build_attributes(k, v, level + 1)
        yield f"{level * INDENT}}}\n"
    else:
        # get type
        ltype = get_type(values)

        # encode values
        if (
            isinstance(values, str)
            or not isinstance(values, Iterable)
            or getattr(values, "shape", None) == ()
        ):
            values = [encode(values)]
        else:
            values = map(encode, values)

        yield f"{level * INDENT}{ltype} {quote(attr)} {', '.join(values)};\n"


def get_type(values):
    """Extract the type of a variable.

    This function tries to determine the DAP type of a Python variable using
    several methods. Returns the DAP type as a string.

    """
    if hasattr(values, "dtype"):
        return NUMPY_TO_DAP2_TYPEMAP[values.dtype.char]
    if isinstance(values, str) or not isinstance(values, Iterable):
        return type_convert(values)
    # if there are several values, they may have different types, so we
    # need to convert all of them and use a precedence table
    types = [type_convert(val) for val in values]
    precedence = ["String", "Float64", "Int32"]
    types.sort(key=precedence.index)
    return types[0]


def type_convert(obj):
    """Map Python objects to the corresponding Opendap types.

    Returns the DAP representation of the type as a string.

    """
    if isinstance(obj, float):
        return "Float64"
    if isinstance(obj, int):
        return "Int32"
    return "String"



"""Test dapclient namespace."""

import unittest

import dapclient


class TestNamespace(unittest.TestCase):

    """Test dapclient namespace."""

    def test_namespace(self):
        """Test the namespace."""
        self.assertEqual(dapclient.__name__, "dapclient")
